# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
[MS-DCOM]

https://learn.microsoft.com/en-us/openspecs/windows_protocols/ms-dcom/4a893f3d-bd29-48cd-9f43-d9777a4415b0
"""

import collections
import uuid

from scapy.config import conf
from scapy.packet import Packet, bind_layers
from scapy.fields import (
    ConditionalField,
    LEIntField,
    LEShortEnumField,
    LEShortField,
    PacketField,
    PacketListField,
    StrNullFieldUtf16,
    UUIDField,
    XStrFixedLenField,
    XShortField,
)
from scapy.layers.dcerpc import (
    NDRFieldListField,
    NDRIntField,
    NDRLongField,
    NDRPacket,
    NDRPacketField,
    NDRFullPointerField,
    NDRConfPacketListField,
    NDRConfFieldListField,
    NDRConfStrLenFieldUtf16,
    NDRConfVarStrNullFieldUtf16,
    NDRShortField,
    NDRSignedIntField,
    NDRSerializeType1PacketField,
    NDRSerializeType1PacketListField,
    ndr_deserialize1,
    find_dcerpc_interface,
    RPC_C_AUTHN,
)
from scapy.layers.msrpce.rpcclient import DCERPC_Client, DCERPC_Transport

from scapy.layers.msrpce.raw.ms_dcom import (
    COMVERSION,
    GUID,
    ServerAlive2_Request,
    MInterfacePointer,
)


def _uid_to_bytes(x, ndrendian="little"):
    if ndrendian == "little":
        return uuid.UUID(x).bytes_le
    elif ndrendian == "big":
        return uuid.UUID(x).bytes
    else:
        raise ValueError("bad ndrendian")


def _uid_from_bytes(x, ndrendian="little"):
    if ndrendian == "little":
        return uuid.UUID(bytes_le=x)
    elif ndrendian == "big":
        return uuid.UUID(bytes=x)
    else:
        raise ValueError("bad ndrendian")


# [MS-DCOM] sect 1.9

CLSID_ActivationContextInfo = uuid.UUID("000001a5-0000-0000-c000-000000000046")
CLSID_ActivationPropertiesIn = uuid.UUID("00000338-0000-0000-c000-000000000046")
CLSID_ActivationPropertiesOut = uuid.UUID("00000339-0000-0000-c000-000000000046")
CLSID_CONTEXT_EXTENSION = uuid.UUID("00000334-0000-0000-c000-000000000046")
CLSID_ContextMarshaler = uuid.UUID("0000033b-0000-0000-c000-000000000046")
CLSID_ERROR_EXTENSION = uuid.UUID("0000031c-0000-0000-c000-000000000046")
CLSID_ErrorObject = uuid.UUID("0000031b-0000-0000-c000-000000000046")
CLSID_InstanceInfo = uuid.UUID("000001ad-0000-0000-c000-000000000046")
CLSID_InstantiationInfo = uuid.UUID("000001ab-0000-0000-c000-000000000046")
CLSID_PropsOutInfo = uuid.UUID("00000339-0000-0000-c000-000000000046")
CLSID_ScmReplyInfo = uuid.UUID("000001b6-0000-0000-c000-000000000046")
CLSID_ScmRequestInfo = uuid.UUID("000001aa-0000-0000-c000-000000000046")
CLSID_SecurityInfo = uuid.UUID("000001a6-0000-0000-c000-000000000046")
CLSID_ServerLocationInfo = uuid.UUID("000001a4-0000-0000-c000-000000000046")
CLSID_SpecialSystemProperties = uuid.UUID("000001b9-0000-0000-c000-000000000046")

# Some special non-interfaces UUIDs:

IID_IActivationPropertiesIn = uuid.UUID("000001A2-0000-0000-C000-000000000046")
IID_IActivationPropertiesOut = uuid.UUID("000001A3-0000-0000-C000-000000000046")
IID_IContext = uuid.UUID("000001c0-0000-0000-C000-000000000046")

# [MS-DCOM] 2.2.22.2.1


class InstantiationInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRPacketField("classId", GUID(), GUID),
        NDRIntField("classCtx", 0),
        NDRIntField("actvflags", 0),
        NDRSignedIntField("fIsSurrogate", 0),
        NDRIntField("cIID", 0),
        NDRIntField("instFlag", 0),
        NDRFullPointerField(
            NDRConfPacketListField(
                "pIID", [GUID()], GUID, count_from=lambda pkt: pkt.cIID
            ),
            deferred=True,
        ),
        NDRIntField("thisSize", 0),
        NDRPacketField("clientCOMVersion", COMVERSION(), COMVERSION),
    ]


# [MS-DCOM] 2.2.22.2.2


class SpecialPropertiesData(NDRPacket):
    ALIGNMENT = (8, 8)
    fields_desc = [
        NDRIntField("dwSessionId", 0),
        NDRSignedIntField("fRemoteThisSessionId", 0),
        NDRSignedIntField("fClientImpersonating", 0),
        NDRSignedIntField("fPartitionIDPresent", 0),
        NDRIntField("dwDefaultAuthnLvl", 0),
        NDRPacketField("guidPartition", GUID(), GUID),
        NDRIntField("dwPRTFlags", 0),
        NDRIntField("dwOrigClsctx", 0),
        NDRIntField("dwFlags", 0),
        NDRIntField("Reserved1", 0),
        NDRLongField("Reserved2", 0),
        NDRFieldListField("Reserved3", [], NDRIntField("", 0), count_from=lambda _: 5),
    ]


# [MS-DCOM] 2.2.22.2.3


class InstanceInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("fileName", ""), deferred=True),
        NDRIntField("mode", 0),
        NDRFullPointerField(
            NDRPacketField("ifdROT", MInterfacePointer(), MInterfacePointer),
            deferred=True,
        ),
        NDRFullPointerField(
            NDRPacketField("ifdStg", MInterfacePointer(), MInterfacePointer),
            deferred=True,
        ),
    ]


# [MS-DCOM] 2.2.22.2.4


class customREMOTE_REQUEST_SCM_INFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("ClientImpLevel", 0),
        NDRShortField("cRequestedProtseqs", 0),
        NDRFullPointerField(
            NDRConfStrLenFieldUtf16(
                "pRequestedProtseqs", "", length_from=lambda pkt: pkt.cRequestedProtseqs
            ),
            deferred=True,
        ),
    ]


class ScmRequestInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(NDRIntField("pdwReserved", 0), deferred=True),
        NDRFullPointerField(
            NDRPacketField(
                "remoteRequest",
                customREMOTE_REQUEST_SCM_INFO(),
                customREMOTE_REQUEST_SCM_INFO,
            ),
            deferred=True,
        ),
    ]


# [MS-DCOM] 2.2.22.2.5


class ActivationContextInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRSignedIntField("clientOK", 0),
        NDRSignedIntField("bReserved1", 0),
        NDRIntField("dwReserved1", 0),
        NDRIntField("dwReserved2", 0),
        NDRFullPointerField(
            NDRPacketField("pIFDClientCtx", MInterfacePointer(), MInterfacePointer),
            deferred=True,
        ),
        NDRFullPointerField(
            NDRPacketField("pIFDPrototypeCtx", MInterfacePointer(), MInterfacePointer),
            deferred=True,
        ),
    ]


# [MS-DCOM] 2.2.22.2.6


class LocationInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("machineName", ""), deferred=True
        ),
        NDRIntField("processId", 0),
        NDRIntField("apartmentId", 0),
        NDRIntField("contextId", 0),
    ]


# [MS-DCOM] 2.2.22.2.7


class COSERVERINFO(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwReserved1", 0),
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("pwszName", ""), deferred=True),
        NDRFullPointerField(NDRIntField("pdwReserved", 0), deferred=True),
        NDRIntField("dwReserved2", 0),
    ]


class SecurityInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("dwAuthnFlags", 0),
        NDRFullPointerField(
            NDRPacketField("pServerInfo", COSERVERINFO(), COSERVERINFO), deferred=True
        ),
        NDRFullPointerField(NDRIntField("pdwReserved", 0), deferred=True),
    ]


# [MS-DCOM] 2.2.22.2.8


class DUALSTRINGARRAY(NDRPacket):
    ALIGNMENT = (4, 8)
    CONFORMANT_COUNT = 1
    fields_desc = [
        NDRShortField("wNumEntries", 0),
        NDRShortField("wSecurityOffset", 0),
        NDRConfStrLenFieldUtf16(
            "aStringArray", "", length_from=lambda pkt: pkt.wNumEntries
        ),
    ]


def _parseStringArray(self):
    """
    Process aStringArray
    """
    str_fld = PacketListField("", [], STRINGBINDING)
    sec_fld = PacketListField("", [], SECURITYBINDING)
    string = str_fld.getfield(self, self.aStringArray[: self.wSecurityOffset * 2])[1]
    secs = sec_fld.getfield(self, self.aStringArray[self.wSecurityOffset * 2 :])[1]
    return string, secs


class customREMOTE_REPLY_SCM_INFO(NDRPacket):
    ALIGNMENT = (8, 8)
    fields_desc = [
        NDRLongField("Oxid", 0),
        NDRFullPointerField(
            NDRPacketField("pdsaOxidBindings", DUALSTRINGARRAY(), DUALSTRINGARRAY),
            deferred=True,
        ),
        NDRPacketField("ipidRemUnknown", GUID(), GUID),
        NDRIntField("authnHint", 0),
        NDRPacketField("serverVersion", COMVERSION(), COMVERSION),
    ]


class ScmReplyInfoData(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(NDRIntField("pdwReserved", 0), deferred=True),
        NDRFullPointerField(
            NDRPacketField(
                "remoteReply",
                customREMOTE_REPLY_SCM_INFO(),
                customREMOTE_REPLY_SCM_INFO,
            ),
            deferred=True,
        ),
    ]


# [MS-DCOM] 2.2.22.2.9


class PropsOutInfo(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("cIfs", 0),
        NDRFullPointerField(
            NDRConfPacketListField(
                "piid", [GUID()], GUID, count_from=lambda pkt: pkt.cIfs
            ),
            deferred=True,
        ),
        NDRFullPointerField(
            NDRConfFieldListField(
                "phresults",
                [],
                NDRSignedIntField("", 0),
                count_from=lambda pkt: pkt.cIfs,
            ),
            deferred=True,
        ),
        NDRFullPointerField(
            NDRConfPacketListField(
                "ppIntfData",
                [MInterfacePointer()],
                MInterfacePointer,
                count_from=lambda pkt: pkt.cIfs,
            ),
            deferred=True,
        ),
    ]


# [MS-DCOM] 2.2.22.1


class CustomHeader(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("totalSize", 0),
        NDRIntField("headerSize", 0),
        NDRIntField("dwReserved", 0),
        NDRIntField("destCtx", 0),
        NDRIntField("cIfs", 0),
        NDRPacketField("classInfoClsid", GUID(), GUID),
        NDRFullPointerField(
            NDRConfPacketListField(
                "pclsid", [GUID()], GUID, count_from=lambda pkt: pkt.cIfs
            ),
            deferred=True,
        ),
        NDRFullPointerField(
            NDRConfFieldListField(
                "pSizes", [], NDRIntField("", 0), count_from=lambda pkt: pkt.cIfs
            ),
            deferred=True,
        ),
        NDRFullPointerField(NDRIntField("pdwReserved", 0), deferred=True),
    ]


class _ActivationPropertiesField(NDRSerializeType1PacketListField):
    def __init__(self, *args, **kwargs):
        kwargs["next_cls_cb"] = self._get_cls_activation
        # kwargs["ptr"] = False
        super(_ActivationPropertiesField, self).__init__(*args, **kwargs)

    def _get_cls_activation(self, pkt, lst, cur, remain):
        pclsid = pkt.CustomHeader.data.pclsid.value.value
        ndrendian = pkt.CustomHeader.data.ndrendian
        i = len(lst) + int(bool(cur))
        if i >= len(pclsid):
            return
        next_uid = _uid_from_bytes(pclsid[i], ndrendian=ndrendian)
        # [MS-DCOM] 1.9
        cls = {
            CLSID_ActivationContextInfo: ActivationContextInfoData,
            CLSID_InstanceInfo: InstanceInfoData,
            CLSID_InstantiationInfo: InstantiationInfoData,
            CLSID_PropsOutInfo: PropsOutInfo,
            CLSID_ScmReplyInfo: ScmReplyInfoData,
            CLSID_ScmRequestInfo: ScmRequestInfoData,
            CLSID_SecurityInfo: SecurityInfoData,
            CLSID_ServerLocationInfo: LocationInfoData,
            CLSID_SpecialSystemProperties: SpecialPropertiesData,
        }[next_uid]
        return lambda x: ndr_deserialize1(x, cls, ndr64=False)


class ActivationPropertiesBlob(Packet):
    fields_desc = [
        LEIntField("dwSize", 0),
        LEIntField("dwReserved", 0),
        NDRSerializeType1PacketField("CustomHeader", CustomHeader(), CustomHeader),
        _ActivationPropertiesField("Property", []),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-DCOM] 2.2.18


class OBJREF(Packet):
    fields_desc = [
        XStrFixedLenField("signature", b"MEOW", length=4),  # :3
        LEIntField("flags", 0x04),
        XStrFixedLenField("iid", IID_IActivationPropertiesIn, length=16),
    ]


# [MS-DCOM] 2.2.18.6


class OBJREF_CUSTOM(Packet):
    fields_desc = [
        UUIDField("clsid", CLSID_ActivationPropertiesIn),
        LEIntField("cbExtension", 0),
        LEIntField("reserved", 0),
        PacketField(
            "pObjectData", ActivationPropertiesBlob(), ActivationPropertiesBlob
        ),
    ]


# [MS-DCOM] 2.2.19.3


class STRINGBINDING(Packet):
    fields_desc = [
        LEShortField("wTowerId", 0),
        StrNullFieldUtf16("aNetworkAddr", ""),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


# [MS-DCOM] 2.2.19.4


class SECURITYBINDING(Packet):
    fields_desc = [
        LEShortEnumField("wAuthnSvc", 0, RPC_C_AUTHN),
        ConditionalField(XShortField("Reserved", 0xFFFF), lambda pkt: pkt.wAuthnSvc),
        ConditionalField(
            StrNullFieldUtf16("aPrincName", ""), lambda pkt: pkt.wAuthnSvc
        ),
    ]

    def default_payload_class(self, payload):
        return conf.padding_layer


bind_layers(OBJREF, OBJREF_CUSTOM, flags=4)


class DCOM_Client(DCERPC_Client):
    """
    A wrapper of DCERPC_Client that adds functions to use COM interfaces.

    In this client, the DCE/RPC is abstracted to allow to focus on the upper
    DCOM one. DCE/RPC interfaces are bound automatically and ORPCTHIS/ORPCTHAT
    automatically added/extracted.

    It also provides common handlers for the few [MS-DCOM] special interfaces.
    """

    def __init__(self, verb=True, **kwargs):
        super(DCOM_Client, self).__init__(
            DCERPC_Transport.NCACN_IP_TCP, ndr64=False, verb=verb, **kwargs
        )

    def connect(self, *args, **kwargs):
        kwargs.setdefault("port", 135)
        super(DCOM_Client, self).connect(*args, **kwargs)

    def ServerAlive2(self):
        """
        Call IObjectExporter::ServerAlive2
        """
        self.bind_or_alter(find_dcerpc_interface("IObjectExporter"))
        resp = self.sr1_req(ServerAlive2_Request(ndr64=False))
        binds, secs = _parseStringArray(resp.ppdsaOrBindings.value)
        DCOMResults = collections.namedtuple("DCOMResults", ["addresses", "ssps"])
        addresses = []
        ssps = []
        for b in binds:
            if b.wTowerId == 0:
                continue
            addresses.append(b.aNetworkAddr)
        for b in secs:
            ssps.append(
                "%s%s"
                % (
                    b.sprintf("%wAuthnSvc%"),
                    b.aPrincName and "%s/" % b.aPrincName or "",
                )
            )
        return DCOMResults(addresses, ssps)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
DCE/RPC client as per [MS-RPCE]
"""

import uuid
import socket

from scapy.config import conf

from scapy.layers.dcerpc import (
    DceRpc5,
    DceRpc5AlterContext,
    DceRpc5AlterContextResp,
    DceRpc5Auth3,
    DceRpc5Bind,
    DceRpc5BindAck,
    DceRpc5BindNak,
    DceRpc5Context,
    DceRpc5Fault,
    DceRpc5Request,
    DceRpc5Response,
    DceRpc5AbstractSyntax,
    DceRpc5TransferSyntax,
    DceRpcSocket,
    DCERPC_Transport,
    find_dcerpc_interface,
    CommonAuthVerifier,
    DCE_C_AUTHN_LEVEL,
    # NDR
    NDRPointer,
    NDRContextHandle,
)
from scapy.layers.gssapi import (
    SSP,
    GSS_S_FAILURE,
    GSS_S_COMPLETE,
    GSS_S_CONTINUE_NEEDED,
    GSS_C_FLAGS,
)
from scapy.layers.smb2 import STATUS_ERREF
from scapy.layers.smbclient import (
    SMB_RPC_SOCKET,
)

# RPC
from scapy.layers.msrpce.ept import (
    ept_map_Request,
    ept_map_Response,
    twr_p_t,
    protocol_tower_t,
    prot_and_addr_t,
    UUID,
)


class DCERPC_Client(object):
    """
    A basic DCE/RPC client

    :param ndr64: Should ask for NDR64 when binding (default False)
    """

    def __init__(self, transport, ndr64=False, ndrendian="little", verb=True, **kwargs):
        self.sock = None
        self.transport = transport
        assert isinstance(
            transport, DCERPC_Transport
        ), "transport must be from DCERPC_Transport"
        self.call_id = 0
        self.cont_id = 0
        self.ndr64 = ndr64
        self.ndrendian = ndrendian
        self.verb = verb
        self.auth_level = kwargs.pop("auth_level", DCE_C_AUTHN_LEVEL.NONE)
        self.auth_context_id = kwargs.pop("auth_context_id", 0)
        self.ssp = kwargs.pop("ssp", None)  # type: SSP
        self.sspcontext = None
        self.dcesockargs = kwargs

    @classmethod
    def from_smblink(cls, smbcli, smb_kwargs={}, **kwargs):
        """
        Build a DCERPC_Client from a SMB_Client.smblink directly
        """
        client = DCERPC_Client(DCERPC_Transport.NCACN_NP, **kwargs)
        sock = client.smbrpcsock = SMB_RPC_SOCKET(smbcli, **smb_kwargs)
        client.sock = DceRpcSocket(
            sock,
            DceRpc5,
            ssp=client.ssp,
            auth_level=client.auth_level,
            auth_context_id=client.auth_context_id,
            **client.dcesockargs,
        )
        return client

    def connect(self, ip, port=None, timeout=5, smb_kwargs={}):
        """
        Initiate a connection
        """
        if port is None:
            if self.transport == DCERPC_Transport.NCACN_IP_TCP:  # IP/TCP
                port = 135
            elif self.transport == DCERPC_Transport.NCACN_NP:  # SMB
                port = 445
            else:
                raise ValueError(
                    "Can't guess the port for transport: %s" % self.transport
                )
        sock = socket.socket()
        sock.settimeout(timeout)
        if self.verb:
            print(
                "\u2503 Connecting to %s on port %s via %s..."
                % (ip, port, repr(self.transport))
            )
        sock.connect((ip, port))
        if self.verb:
            print(
                conf.color_theme.green(
                    "\u2514 Connected from %s" % repr(sock.getsockname())
                )
            )
        if self.transport == DCERPC_Transport.NCACN_NP:  # SMB
            # We pack the socket into a SMB_RPC_SOCKET
            sock = self.smbrpcsock = SMB_RPC_SOCKET.from_tcpsock(
                sock, ssp=self.ssp, **smb_kwargs
            )
            self.sock = DceRpcSocket(sock, DceRpc5, **self.dcesockargs)
        elif self.transport == DCERPC_Transport.NCACN_IP_TCP:
            self.sock = DceRpcSocket(
                sock,
                DceRpc5,
                ssp=self.ssp,
                auth_level=self.auth_level,
                auth_context_id=self.auth_context_id,
                **self.dcesockargs,
            )

    def close(self):
        if self.verb:
            print("X Connection closed\n")
        self.sock.close()

    def sr1(self, pkt, **kwargs):
        self.call_id += 1
        pkt = (
            DceRpc5(
                call_id=self.call_id,
                pfc_flags="PFC_FIRST_FRAG+PFC_LAST_FRAG",
                endian=self.ndrendian,
                auth_verifier=kwargs.pop("auth_verifier", None),
            )
            / pkt
        )
        if "pfc_flags" in kwargs:
            pkt.pfc_flags = kwargs.pop("pfc_flags")
        return self.sock.sr1(pkt, verbose=0, **kwargs)

    def send(self, pkt, **kwargs):
        self.call_id += 1
        pkt = (
            DceRpc5(
                call_id=self.call_id,
                pfc_flags="PFC_FIRST_FRAG+PFC_LAST_FRAG",
                endian=self.ndrendian,
                auth_verifier=kwargs.pop("auth_verifier", None),
            )
            / pkt
        )
        if "pfc_flags" in kwargs:
            pkt.pfc_flags = kwargs.pop("pfc_flags")
        return self.sock.send(pkt, **kwargs)

    def sr1_req(self, pkt, **kwargs):
        if self.verb:
            print(conf.color_theme.opening(">> REQUEST: %s" % pkt.__class__.__name__))
        # Send/receive
        resp = self.sr1(
            DceRpc5Request(cont_id=self.cont_id, alloc_hint=len(pkt)) / pkt,
            **kwargs,
        )
        if DceRpc5Response in resp:
            if self.verb:
                print(
                    conf.color_theme.success(
                        "<< RESPONSE: %s"
                        % (resp[DceRpc5Response].payload.__class__.__name__)
                    )
                )
            return resp[DceRpc5Response].payload
        else:
            if self.verb:
                if DceRpc5Fault in resp:
                    if resp[DceRpc5Fault].payload and not isinstance(
                        resp[DceRpc5Fault].payload, conf.raw_layer
                    ):
                        resp[DceRpc5Fault].payload.show()
                    if resp.status == 0x00000005:
                        print(conf.color_theme.fail("! nca_s_fault_access_denied"))
                    elif resp.status == 0x00000721:
                        print(
                            conf.color_theme.fail(
                                "! nca_s_fault_sec_pkg_error "
                                "(error in checksum/encryption)"
                            )
                        )
                    else:
                        print(
                            conf.color_theme.fail(
                                "! %s" % STATUS_ERREF.get(resp.status, "Failure")
                            )
                        )
                        resp.show()
                return
            return resp

    def get_bind_context(self, interface):
        return [
            DceRpc5Context(
                cont_id=0,
                abstract_syntax=DceRpc5AbstractSyntax(
                    if_uuid=interface.uuid,
                    if_version=interface.if_version,
                ),
                transfer_syntaxes=[
                    DceRpc5TransferSyntax(
                        # NDR 2.0 32-bit
                        if_uuid="NDR 2.0",
                        if_version=2,
                    )
                ],
            ),
        ] + (
            [
                DceRpc5Context(
                    cont_id=1,
                    abstract_syntax=DceRpc5AbstractSyntax(
                        if_uuid=interface.uuid,
                        if_version=interface.if_version,
                    ),
                    transfer_syntaxes=[
                        DceRpc5TransferSyntax(
                            # NDR64
                            if_uuid="NDR64",
                            if_version=1,
                        )
                    ],
                ),
                DceRpc5Context(
                    cont_id=2,
                    abstract_syntax=DceRpc5AbstractSyntax(
                        if_uuid=interface.uuid,
                        if_version=interface.if_version,
                    ),
                    transfer_syntaxes=[
                        DceRpc5TransferSyntax(
                            if_uuid=uuid.UUID("6cb71c2c-9812-4540-0300-000000000000"),
                            if_version=1,
                        )
                    ],
                ),
            ]
            if self.ndr64
            else []
        )

    def _bind(self, interface, reqcls, respcls):
        # Build a security context: [MS-RPCE] 3.3.1.5.2
        if self.verb:
            print(
                conf.color_theme.opening(
                    ">> %s on %s" % (reqcls.__name__, interface)
                    + (" (with %s)" % self.ssp.__class__.__name__ if self.ssp else "")
                )
            )
        if not self.ssp or (
            self.transport == DCERPC_Transport.NCACN_NP
            and self.auth_level < DCE_C_AUTHN_LEVEL.PKT_INTEGRITY
        ):
            # NCACN_NP = SMB without INTEGRITY/PRIVACY does not bind the RPC securely,
            # again as it has already authenticated during the SMB Session Setup
            resp = self.sr1(
                reqcls(context_elem=self.get_bind_context(interface)),
                auth_verifier=None,
            )
            status = GSS_S_COMPLETE
        else:
            # Perform authentication
            self.sspcontext, token, status = self.ssp.GSS_Init_sec_context(
                self.sspcontext,
                req_flags=(
                    # SSPs need to be instantiated with some special flags
                    # for DCE/RPC usages.
                    GSS_C_FLAGS.GSS_C_DCE_STYLE
                    | GSS_C_FLAGS.GSS_C_REPLAY_FLAG
                    | GSS_C_FLAGS.GSS_C_SEQUENCE_FLAG
                    | GSS_C_FLAGS.GSS_C_MUTUAL_FLAG
                    | (
                        GSS_C_FLAGS.GSS_C_INTEG_FLAG
                        if self.auth_level >= DCE_C_AUTHN_LEVEL.PKT_INTEGRITY
                        else 0
                    )
                    | (
                        GSS_C_FLAGS.GSS_C_CONF_FLAG
                        if self.auth_level >= DCE_C_AUTHN_LEVEL.PKT_PRIVACY
                        else 0
                    )
                ),
            )
            if status not in [GSS_S_CONTINUE_NEEDED, GSS_S_COMPLETE]:
                # Authentication failed.
                self.sspcontext.clifailure()
                return False
            resp = self.sr1(
                reqcls(context_elem=self.get_bind_context(interface)),
                auth_verifier=(
                    None
                    if not self.sspcontext
                    else CommonAuthVerifier(
                        auth_type=self.ssp.auth_type,
                        auth_level=self.auth_level,
                        auth_context_id=self.auth_context_id,
                        auth_value=token,
                    )
                ),
                pfc_flags=(
                    "PFC_FIRST_FRAG+PFC_LAST_FRAG"
                    + (
                        # If the SSP supports "Header Signing", advertise it
                        "+PFC_SUPPORT_HEADER_SIGN"
                        if self.ssp is not None
                        and self.sock.session.support_header_signing
                        else ""
                    )
                ),
            )
            if respcls not in resp:
                token = None
                status = GSS_S_FAILURE
            else:
                # Call the underlying SSP
                self.sspcontext, token, status = self.ssp.GSS_Init_sec_context(
                    self.sspcontext, val=resp.auth_verifier.auth_value
                )
            if status in [GSS_S_CONTINUE_NEEDED, GSS_S_COMPLETE]:
                # Authentication should continue
                if token and self.ssp.LegsAmount(self.sspcontext) % 2 == 1:
                    # AUTH 3 for certain SSPs (e.g. NTLM)
                    # "The server MUST NOT respond to an rpc_auth_3 PDU"
                    self.send(
                        DceRpc5Auth3(),
                        auth_verifier=CommonAuthVerifier(
                            auth_type=self.ssp.auth_type,
                            auth_level=self.auth_level,
                            auth_context_id=self.auth_context_id,
                            auth_value=token,
                        ),
                    )
                    status = GSS_S_COMPLETE
                else:
                    # Authentication can continue in two ways:
                    # - through DceRpc5Auth3 (e.g. NTLM)
                    # - through DceRpc5AlterContext (e.g. Kerberos)
                    while token:
                        respcls = DceRpc5AlterContextResp
                        resp = self.sr1(
                            DceRpc5AlterContext(
                                context_elem=self.get_bind_context(interface)
                            ),
                            auth_verifier=CommonAuthVerifier(
                                auth_type=self.ssp.auth_type,
                                auth_level=self.auth_level,
                                auth_context_id=self.auth_context_id,
                                auth_value=token,
                            ),
                        )
                        if respcls not in resp:
                            status = GSS_S_FAILURE
                            break
                        if resp.auth_verifier is None:
                            status = GSS_S_COMPLETE
                            break
                        self.sspcontext, token, status = self.ssp.GSS_Init_sec_context(
                            self.sspcontext, val=resp.auth_verifier.auth_value
                        )
        # Check context acceptance
        if (
            status == GSS_S_COMPLETE
            and respcls in resp
            and any(x.result == 0 for x in resp.results[: int(self.ndr64) + 1])
        ):
            self.call_id = 0  # reset call id
            port = resp.sec_addr.port_spec.decode()
            ndr = self.sock.session.ndr64 and "NDR64" or "NDR32"
            self.cont_id = int(self.sock.session.ndr64)  # ctx 0 for NDR32, 1 for NDR64
            if self.verb:
                print(
                    conf.color_theme.success(
                        f"<< {respcls.__name__} port '{port}' using {ndr}"
                    )
                )
            self.sock.session.sspcontext = self.sspcontext
            return True
        else:
            if self.verb:
                if DceRpc5BindNak in resp:
                    err_msg = resp.sprintf(
                        "reject_reason: %DceRpc5BindNak.provider_reject_reason%"
                    )
                    print(conf.color_theme.fail("! Bind_nak (%s)" % err_msg))
                    if DceRpc5BindNak in resp:
                        if resp[DceRpc5BindNak].payload and not isinstance(
                            resp[DceRpc5BindNak].payload, conf.raw_layer
                        ):
                            resp[DceRpc5BindNak].payload.show()
                elif DceRpc5Fault in resp:
                    if resp.status == 0x00000005:
                        print(conf.color_theme.fail("! nca_s_fault_access_denied"))
                    elif resp.status == 0x00000721:
                        print(
                            conf.color_theme.fail(
                                "! nca_s_fault_sec_pkg_error "
                                "(error in checksum/encryption)"
                            )
                        )
                    else:
                        print(
                            conf.color_theme.fail(
                                "! %s" % STATUS_ERREF.get(resp.status, "Failure")
                            )
                        )
                        resp.show()
                    if DceRpc5Fault in resp:
                        if resp[DceRpc5Fault].payload and not isinstance(
                            resp[DceRpc5Fault].payload, conf.raw_layer
                        ):
                            resp[DceRpc5Fault].payload.show()
                else:
                    print(conf.color_theme.fail("! Failure"))
                    resp.show()
            return False

    def bind(self, interface):
        """
        Bind the client to an interface
        """
        return self._bind(interface, DceRpc5Bind, DceRpc5BindAck)

    def alter_context(self, interface):
        """
        Alter context: post-bind context negotiation
        """
        return self._bind(interface, DceRpc5AlterContext, DceRpc5AlterContextResp)

    def bind_or_alter(self, interface):
        """
        Bind the client to an interface or alter the context if already bound
        """
        if not self.sock.session.rpc_bind_interface:
            # No interface is bound
            self.bind(interface)
        else:
            # An interface is already bound
            self.alter_context(interface)

    def open_smbpipe(self, name):
        """
        Open a certain filehandle with the SMB automaton
        """
        self.ipc_tid = self.smbrpcsock.tree_connect("IPC$")
        self.smbrpcsock.open_pipe(name)

    def close_smbpipe(self):
        """
        Close the previously opened pipe
        """
        self.smbrpcsock.set_TID(self.ipc_tid)
        self.smbrpcsock.close_pipe()
        self.smbrpcsock.tree_disconnect()

    def connect_and_bind(
        self,
        ip,
        interface,
        port=None,
        smb_kwargs={},
    ):
        """
        Asks the Endpoint Mapper what address to use to connect to the interface,
        then uses connect() followed by a bind()
        """
        if self.transport == DCERPC_Transport.NCACN_IP_TCP:
            # IP/TCP
            # 1. ask the endpoint mapper (port 135) for the IP:PORT
            endpoints = get_endpoint(
                ip,
                interface,
                ndrendian=self.ndrendian,
                verb=self.verb,
            )
            if endpoints:
                ip, port = endpoints[0]
            else:
                return
            # 2. Connect to that IP:PORT
            self.connect(ip, port=port)
        elif self.transport == DCERPC_Transport.NCACN_NP:
            # SMB
            # 1. ask the endpoint mapper (over SMB) for the namedpipe
            endpoints = get_endpoint(
                ip,
                interface,
                transport=self.transport,
                ndrendian=self.ndrendian,
                verb=self.verb,
                smb_kwargs=smb_kwargs,
            )
            if endpoints:
                pipename = endpoints[0].lstrip("\\pipe\\")
            else:
                return
            # 2. connect to the SMB server
            self.connect(ip, port=port, smb_kwargs=smb_kwargs)
            # 3. open the new named pipe
            self.open_smbpipe(pipename)
        # Bind in RPC
        self.bind(interface)

    def epm_map(self, interface):
        """
        Calls ept_map (the EndPoint Manager)
        """
        if self.ndr64:
            ndr_uuid = "NDR64"
            ndr_version = 1
        else:
            ndr_uuid = "NDR 2.0"
            ndr_version = 2
        pkt = self.sr1_req(
            ept_map_Request(
                obj=NDRPointer(
                    referent_id=1,
                    value=UUID(
                        Data1=0,
                        Data2=0,
                        Data3=0,
                        Data4=None,
                    ),
                ),
                map_tower=NDRPointer(
                    referent_id=2,
                    value=twr_p_t(
                        tower_octet_string=bytes(
                            protocol_tower_t(
                                floors=[
                                    prot_and_addr_t(
                                        lhs_length=19,
                                        protocol_identifier=0xD,
                                        uuid=interface.uuid,
                                        version=interface.major_version,
                                        rhs_length=2,
                                        rhs=interface.minor_version,
                                    ),
                                    prot_and_addr_t(
                                        lhs_length=19,
                                        protocol_identifier=0xD,
                                        uuid=ndr_uuid,
                                        version=ndr_version,
                                        rhs_length=2,
                                        rhs=0,
                                    ),
                                    prot_and_addr_t(
                                        lhs_length=1,
                                        protocol_identifier="RPC connection-oriented protocol",  # noqa: E501
                                        rhs_length=2,
                                        rhs=0,
                                    ),
                                    {
                                        DCERPC_Transport.NCACN_IP_TCP: (
                                            prot_and_addr_t(
                                                lhs_length=1,
                                                protocol_identifier="NCACN_IP_TCP",
                                                rhs_length=2,
                                                rhs=135,
                                            )
                                        ),
                                        DCERPC_Transport.NCACN_NP: (
                                            prot_and_addr_t(
                                                lhs_length=1,
                                                protocol_identifier="NCACN_NP",
                                                rhs_length=2,
                                                rhs=b"0\x00",
                                            )
                                        ),
                                    }[self.transport],
                                    {
                                        DCERPC_Transport.NCACN_IP_TCP: (
                                            prot_and_addr_t(
                                                lhs_length=1,
                                                protocol_identifier="IP",
                                                rhs_length=4,
                                                rhs="0.0.0.0",
                                            )
                                        ),
                                        DCERPC_Transport.NCACN_NP: (
                                            prot_and_addr_t(
                                                lhs_length=1,
                                                protocol_identifier="NCACN_NB",
                                                rhs_length=10,
                                                rhs=b"127.0.0.1\x00",
                                            )
                                        ),
                                    }[self.transport],
                                ],
                            )
                        ),
                    ),
                ),
                entry_handle=NDRContextHandle(
                    attributes=0,
                    uuid=b"\x00" * 16,
                ),
                max_towers=500,
                ndr64=self.ndr64,
                ndrendian=self.ndrendian,
            )
        )
        if pkt and ept_map_Response in pkt:
            status = pkt[ept_map_Response].status
            # [MS-RPCE] sect 2.2.1.2.5
            if status == 0x00000000:
                towers = [
                    protocol_tower_t(x.value.tower_octet_string)
                    for x in pkt[ept_map_Response].ITowers.value[0].value
                ]
                # Let's do some checks to know we know what we're doing
                endpoints = []
                for t in towers:
                    if t.floors[0].uuid != interface.uuid:
                        if self.verb:
                            print(
                                conf.color_theme.fail(
                                    "! Server answered with a different interface."
                                )
                            )
                        raise ValueError
                    if t.floors[1].sprintf("%uuid%") != ndr_uuid:
                        if self.verb:
                            print(
                                conf.color_theme.fail(
                                    "! Server answered with a different NDR version."
                                )
                            )
                        raise ValueError
                    if self.transport == DCERPC_Transport.NCACN_IP_TCP:
                        endpoints.append((t.floors[4].rhs, t.floors[3].rhs))
                    elif self.transport == DCERPC_Transport.NCACN_NP:
                        endpoints.append(t.floors[3].rhs.rstrip(b"\x00").decode())
                return endpoints
            elif status == 0x16C9A0D6:
                if self.verb:
                    pkt.show()
                    print(
                        conf.color_theme.fail(
                            "! Server errored: 'There are no elements that satisfy"
                            " the specified search criteria'."
                        )
                    )
                raise ValueError
        print(conf.color_theme.fail("! Failure."))
        if pkt:
            pkt.show()
        raise ValueError("EPM Map failed")


def get_endpoint(
    ip,
    interface,
    transport=DCERPC_Transport.NCACN_IP_TCP,
    ndrendian="little",
    verb=True,
    smb_kwargs={},
):
    """
    Call the endpoint mapper on a remote IP to find an interface

    :param ip:
    :param interface:
    :param mode:
    :param verb:

    :return: a list of connection tuples for this interface
    """
    client = DCERPC_Client(
        transport,
        ndr64=False,
        ndrendian=ndrendian,
        verb=verb,
    )  # EPM only works with NDR32
    client.connect(ip, smb_kwargs=smb_kwargs)
    if transport == DCERPC_Transport.NCACN_NP:  # SMB
        client.open_smbpipe("epmapper")
    client.bind(find_dcerpc_interface("ept"))
    endpoints = client.epm_map(interface)
    client.close()
    return endpoints



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter <gabriel[]potter[]fr>

"""
SPNEGO

Implements parts of:

- GSSAPI SPNEGO: RFC4178 > RFC2478
- GSSAPI SPNEGO NEGOEX: [MS-NEGOEX]

.. note::
    You will find more complete documentation for this layer over at
    `GSSAPI <https://scapy.readthedocs.io/en/latest/layers/gssapi.html#spnego>`_
"""

import struct
from uuid import UUID

from scapy.asn1.asn1 import (
    ASN1_OID,
    ASN1_STRING,
    ASN1_Codecs,
)
from scapy.asn1.mib import conf  # loads conf.mib
from scapy.asn1fields import (
    ASN1F_CHOICE,
    ASN1F_ENUMERATED,
    ASN1F_FLAGS,
    ASN1F_GENERAL_STRING,
    ASN1F_OID,
    ASN1F_PACKET,
    ASN1F_SEQUENCE,
    ASN1F_SEQUENCE_OF,
    ASN1F_STRING,
    ASN1F_optional,
)
from scapy.asn1packet import ASN1_Packet
from scapy.fields import (
    FieldListField,
    LEIntEnumField,
    LEIntField,
    LELongEnumField,
    LELongField,
    LEShortField,
    MultipleTypeField,
    PacketField,
    PacketListField,
    StrField,
    StrFixedLenField,
    UUIDEnumField,
    UUIDField,
    XStrFixedLenField,
    XStrLenField,
)
from scapy.packet import Packet, bind_layers

from scapy.layers.gssapi import (
    GSSAPI_BLOB,
    GSSAPI_BLOB_SIGNATURE,
    GSS_C_FLAGS,
    GSS_S_BAD_MECH,
    GSS_S_COMPLETE,
    GSS_S_CONTINUE_NEEDED,
    SSP,
    _GSSAPI_OIDS,
    _GSSAPI_SIGNATURE_OIDS,
)

# SSP Providers
from scapy.layers.kerberos import (
    Kerberos,
)
from scapy.layers.ntlm import (
    NEGOEX_EXCHANGE_NTLM,
    NTLM_Header,
    _NTLMPayloadField,
    _NTLMPayloadPacket,
)

# Typing imports
from typing import (
    Dict,
    Optional,
    Tuple,
)

# SPNEGO negTokenInit
# https://datatracker.ietf.org/doc/html/rfc4178#section-4.2.1


class SPNEGO_MechType(ASN1_Packet):
    ASN1_codec = ASN1_Codecs.BER
    ASN1_root = ASN1F_OID("oid", None)


class SPNEGO_MechTypes(ASN1_Packet):
    ASN1_codec = ASN1_Codecs.BER
    ASN1_root = ASN1F_SEQUENCE_OF("mechTypes", None, SPNEGO_MechType)


class SPNEGO_MechListMIC(ASN1_Packet):
    ASN1_codec = ASN1_Codecs.BER
    ASN1_root = ASN1F_STRING("value", "")


_mechDissector = {
    "1.3.6.1.4.1.311.2.2.10": NTLM_Header,  # NTLM
    "1.2.840.48018.1.2.2": Kerberos,  # MS KRB5 - Microsoft Kerberos 5
    "1.2.840.113554.1.2.2": Kerberos,  # Kerberos 5
}


class _SPNEGO_Token_Field(ASN1F_STRING):
    def i2m(self, pkt, x):
        if x is None:
            x = b""
        return super(_SPNEGO_Token_Field, self).i2m(pkt, bytes(x))

    def m2i(self, pkt, s):
        dat, r = super(_SPNEGO_Token_Field, self).m2i(pkt, s)
        if isinstance(pkt.underlayer, SPNEGO_negTokenInit):
            types = pkt.underlayer.mechTypes
        elif isinstance(pkt.underlayer, SPNEGO_negTokenResp):
            types = [pkt.underlayer.supportedMech]
        if types and types[0] and types[0].oid.val in _mechDissector:
            return _mechDissector[types[0].oid.val](dat.val), r
        return dat, r


class SPNEGO_Token(ASN1_Packet):
    ASN1_codec = ASN1_Codecs.BER
    ASN1_root = _SPNEGO_Token_Field("value", None)


_ContextFlags = [
    "delegFlag",
    "mutualFlag",
    "replayFlag",
    "sequenceFlag",
    "superseded",
    "anonFlag",
    "confFlag",
    "integFlag",
]


class SPNEGO_negHints(ASN1_Packet):
    # [MS-SPNG] 2.2.1
    ASN1_codec = ASN1_Codecs.BER
    ASN1_root = ASN1F_SEQUENCE(
        ASN1F_optional(
            ASN1F_GENERAL_STRING(
                "hintName", "not_defined_in_RFC4178@please_ignore", explicit_tag=0xA0
            ),
        ),
        ASN1F_optional(
            ASN1F_GENERAL_STRING("hintAddress", None, explicit_tag=0xA1),
        ),
    )


class SPNEGO_negTokenInit(ASN1_Packet):
    ASN1_codec = ASN1_Codecs.BER
    ASN1_root = ASN1F_SEQUENCE(
        ASN1F_optional(
            ASN1F_SEQUENCE_OF("mechTypes", None, SPNEGO_MechType, explicit_tag=0xA0)
        ),
        ASN1F_optional(ASN1F_FLAGS("reqFlags", None, _ContextFlags, implicit_tag=0x81)),
        ASN1F_optional(
            ASN1F_PACKET("mechToken", None, SPNEGO_Token, explicit_tag=0xA2)
        ),
        # [MS-SPNG] flavor !
        ASN1F_optional(
            ASN1F_PACKET("negHints", None, SPNEGO_negHints, explicit_tag=0xA3)
        ),
        ASN1F_optional(
            ASN1F_PACKET("mechListMIC", None, SPNEGO_MechListMIC, explicit_tag=0xA4)
        ),
        # Compat with RFC 4178's SPNEGO_negTokenInit
        ASN1F_optional(
            ASN1F_PACKET("_mechListMIC", None, SPNEGO_MechListMIC, explicit_tag=0xA3)
        ),
    )


# SPNEGO negTokenTarg
# https://datatracker.ietf.org/doc/html/rfc4178#section-4.2.2


class SPNEGO_negTokenResp(ASN1_Packet):
    ASN1_codec = ASN1_Codecs.BER
    ASN1_root = ASN1F_SEQUENCE(
        ASN1F_optional(
            ASN1F_ENUMERATED(
                "negResult",
                0,
                {
                    0: "accept-completed",
                    1: "accept-incomplete",
                    2: "reject",
                    3: "request-mic",
                },
                explicit_tag=0xA0,
            ),
        ),
        ASN1F_optional(
            ASN1F_PACKET(
                "supportedMech", SPNEGO_MechType(), SPNEGO_MechType, explicit_tag=0xA1
            ),
        ),
        ASN1F_optional(
            ASN1F_PACKET("responseToken", None, SPNEGO_Token, explicit_tag=0xA2)
        ),
        ASN1F_optional(
            ASN1F_PACKET("mechListMIC", None, SPNEGO_MechListMIC, explicit_tag=0xA3)
        ),
    )


class SPNEGO_negToken(ASN1_Packet):
    ASN1_codec = ASN1_Codecs.BER
    ASN1_root = ASN1F_CHOICE(
        "token",
        SPNEGO_negTokenInit(),
        ASN1F_PACKET(
            "negTokenInit",
            SPNEGO_negTokenInit(),
            SPNEGO_negTokenInit,
            explicit_tag=0xA0,
        ),
        ASN1F_PACKET(
            "negTokenResp",
            SPNEGO_negTokenResp(),
            SPNEGO_negTokenResp,
            explicit_tag=0xA1,
        ),
    )


# Register for the GSS API Blob

_GSSAPI_OIDS["1.3.6.1.5.5.2"] = SPNEGO_negToken
_GSSAPI_SIGNATURE_OIDS["1.3.6.1.5.5.2"] = SPNEGO_negToken


def mechListMIC(oids):
    """
    Implementation of RFC 4178 - Appendix D. mechListMIC Computation
    """
    return bytes(SPNEGO_MechTypes(mechTypes=oids))


# NEGOEX
# https://docs.microsoft.com/en-us/openspecs/windows_protocols/ms-negoex/0ad7a003-ab56-4839-a204-b555ca6759a2


_NEGOEX_AUTH_SCHEMES = {
    # Reversed. Is there any doc related to this?
    # The NEGOEX doc is very ellusive
    UUID("5c33530d-eaf9-0d4d-b2ec-4ae3786ec308"): "UUID('[NTLM-UUID]')",
}


class NEGOEX_MESSAGE_HEADER(Packet):
    fields_desc = [
        StrFixedLenField("Signature", "NEGOEXTS", length=8),
        LEIntEnumField(
            "MessageType",
            0,
            {
                0x0: "INITIATOR_NEGO",
                0x01: "ACCEPTOR_NEGO",
                0x02: "INITIATOR_META_DATA",
                0x03: "ACCEPTOR_META_DATA",
                0x04: "CHALLENGE",
                0x05: "AP_REQUEST",
                0x06: "VERIFY",
                0x07: "ALERT",
            },
        ),
        LEIntField("SequenceNum", 0),
        LEIntField("cbHeaderLength", None),
        LEIntField("cbMessageLength", None),
        UUIDField("ConversationId", None),
    ]

    def post_build(self, pkt, pay):
        if self.cbHeaderLength is None:
            pkt = pkt[16:] + struct.pack("<I", len(pkt)) + pkt[20:]
        if self.cbMessageLength is None:
            pkt = pkt[20:] + struct.pack("<I", len(pkt) + len(pay)) + pkt[24:]
        return pkt + pay


def _NEGOEX_post_build(self, p, pay_offset, fields):
    # type: (Packet, bytes, int, Dict[str, Tuple[str, int]]) -> bytes
    """Util function to build the offset and populate the lengths"""
    for field_name, value in self.fields["Payload"]:
        length = self.get_field("Payload").fields_map[field_name].i2len(self, value)
        count = self.get_field("Payload").fields_map[field_name].i2count(self, value)
        offset = fields[field_name]
        # Offset
        if self.getfieldval(field_name + "BufferOffset") is None:
            p = p[:offset] + struct.pack("<I", pay_offset) + p[offset + 4 :]
        # Count
        if self.getfieldval(field_name + "Count") is None:
            p = p[: offset + 4] + struct.pack("<H", count) + p[offset + 6 :]
        pay_offset += length
    return p


class NEGOEX_BYTE_VECTOR(Packet):
    fields_desc = [
        LEIntField("ByteArrayBufferOffset", 0),
        LEIntField("ByteArrayLength", 0),
    ]

    def guess_payload_class(self, payload):
        return conf.padding_layer


class NEGOEX_EXTENSION_VECTOR(Packet):
    fields_desc = [
        LELongField("ExtensionArrayOffset", 0),
        LEShortField("ExtensionCount", 0),
    ]


class NEGOEX_NEGO_MESSAGE(_NTLMPayloadPacket):
    OFFSET = 92
    show_indent = 0
    fields_desc = [
        NEGOEX_MESSAGE_HEADER,
        XStrFixedLenField("Random", b"", length=32),
        LELongField("ProtocolVersion", 0),
        LEIntField("AuthSchemeBufferOffset", None),
        LEShortField("AuthSchemeCount", None),
        LEIntField("ExtensionBufferOffset", None),
        LEShortField("ExtensionCount", None),
        # Payload
        _NTLMPayloadField(
            "Payload",
            OFFSET,
            [
                FieldListField(
                    "AuthScheme",
                    [],
                    UUIDEnumField("", None, _NEGOEX_AUTH_SCHEMES),
                    count_from=lambda pkt: pkt.AuthSchemeCount,
                ),
                PacketListField(
                    "Extension",
                    [],
                    NEGOEX_EXTENSION_VECTOR,
                    count_from=lambda pkt: pkt.ExtensionCount,
                ),
            ],
            length_from=lambda pkt: pkt.cbMessageLength - 92,
        ),
        # TODO: dissect extensions
    ]

    def post_build(self, pkt, pay):
        # type: (bytes, bytes) -> bytes
        return (
            _NEGOEX_post_build(
                self,
                pkt,
                self.OFFSET,
                {
                    "AuthScheme": 96,
                    "Extension": 102,
                },
            )
            + pay
        )

    @classmethod
    def dispatch_hook(cls, _pkt=None, *args, **kargs):
        if _pkt and len(_pkt) >= 12:
            MessageType = struct.unpack("<I", _pkt[8:12])[0]
            if MessageType in [0, 1]:
                return NEGOEX_NEGO_MESSAGE
            elif MessageType in [2, 3]:
                return NEGOEX_EXCHANGE_MESSAGE
        return cls


# RFC3961
_checksum_types = {
    1: "CRC32",
    2: "RSA-MD4",
    3: "RSA-MD4-DES",
    4: "DES-MAC",
    5: "DES-MAC-K",
    6: "RSA-MDA-DES-K",
    7: "RSA-MD5",
    8: "RSA-MD5-DES",
    9: "RSA-MD5-DES3",
    10: "SHA1",
    12: "HMAC-SHA1-DES3-KD",
    13: "HMAC-SHA1-DES3",
    14: "SHA1",
    15: "HMAC-SHA1-96-AES128",
    16: "HMAC-SHA1-96-AES256",
}


def _checksum_size(pkt):
    if pkt.ChecksumType == 1:
        return 4
    elif pkt.ChecksumType in [2, 4, 6, 7]:
        return 16
    elif pkt.ChecksumType in [3, 8, 9]:
        return 24
    elif pkt.ChecksumType == 5:
        return 8
    elif pkt.ChecksumType in [10, 12, 13, 14, 15, 16]:
        return 20
    return 0


class NEGOEX_CHECKSUM(Packet):
    fields_desc = [
        LELongField("cbHeaderLength", 20),
        LELongEnumField("ChecksumScheme", 1, {1: "CHECKSUM_SCHEME_RFC3961"}),
        LELongEnumField("ChecksumType", None, _checksum_types),
        XStrLenField("ChecksumValue", b"", length_from=_checksum_size),
    ]


class NEGOEX_EXCHANGE_MESSAGE(_NTLMPayloadPacket):
    OFFSET = 64
    show_indent = 0
    fields_desc = [
        NEGOEX_MESSAGE_HEADER,
        UUIDEnumField("AuthScheme", None, _NEGOEX_AUTH_SCHEMES),
        LEIntField("ExchangeBufferOffset", 0),
        LEIntField("ExchangeLen", 0),
        _NTLMPayloadField(
            "Payload",
            OFFSET,
            [
                # The NEGOEX doc mentions the following blob as as an
                # "opaque handshake for the client authentication scheme".
                # NEGOEX_EXCHANGE_NTLM is a reversed interpretation, and is
                # probably not accurate.
                MultipleTypeField(
                    [
                        (
                            PacketField("Exchange", None, NEGOEX_EXCHANGE_NTLM),
                            lambda pkt: pkt.AuthScheme
                            == UUID("5c33530d-eaf9-0d4d-b2ec-4ae3786ec308"),
                        ),
                    ],
                    StrField("Exchange", b""),
                )
            ],
            length_from=lambda pkt: pkt.cbMessageLength - pkt.cbHeaderLength,
        ),
    ]


class NEGOEX_VERIFY_MESSAGE(Packet):
    show_indent = 0
    fields_desc = [
        NEGOEX_MESSAGE_HEADER,
        UUIDEnumField("AuthScheme", None, _NEGOEX_AUTH_SCHEMES),
        PacketField("Checksum", NEGOEX_CHECKSUM(), NEGOEX_CHECKSUM),
    ]


bind_layers(NEGOEX_NEGO_MESSAGE, NEGOEX_NEGO_MESSAGE)


_mechDissector["1.3.6.1.4.1.311.2.2.30"] = NEGOEX_NEGO_MESSAGE

# -- SSP


class SPNEGOSSP(SSP):
    """
    The SPNEGO SSP

    :param ssps: a dict with keys being the SSP class, and the value being a
                 dictionary of the keyword arguments to pass it on init.

    Example::

        from scapy.layers.ntlm import NTLMSSP
        from scapy.layers.kerberos import KerberosSSP
        from scapy.layers.spnego import SPNEGOSSP
        from scapy.layers.smbserver import smbserver
        from scapy.libs.rfc3961 import Encryption, Key

        ssp = SPNEGOSSP([
            NTLMSSP(
                IDENTITIES={
                    "User1": MD4le("Password1"),
                    "Administrator": MD4le("Password123!"),
                }
            ),
            KerberosSSP(
                SPN="cifs/server2.domain.local",
                KEY=Key(
                    Encryption.AES256,
                    key=hex_bytes("5e9255c907b2f7e969ddad816eabbec8f1f7a387c7194ecc98b827bdc9421c2b")
                )
            )
        ])
        smbserver(ssp=ssp)
    """

    __slots__ = [
        "supported_ssps",
        "force_supported_mechtypes",
    ]
    auth_type = 0x09

    class STATE(SSP.STATE):
        FIRST = 1
        CHANGESSP = 2
        NORMAL = 3

    class CONTEXT(SSP.CONTEXT):
        __slots__ = [
            "supported_mechtypes",
            "requested_mechtypes",
            "req_flags",
            "negotiated_mechtype",
            "first_choice",
            "sub_context",
            "ssp",
        ]

        def __init__(
            self, supported_ssps, req_flags=None, force_supported_mechtypes=None
        ):
            self.state = SPNEGOSSP.STATE.FIRST
            self.requested_mechtypes = None
            self.req_flags = req_flags
            self.first_choice = True
            self.negotiated_mechtype = None
            self.sub_context = None
            self.ssp = None
            if force_supported_mechtypes is None:
                self.supported_mechtypes = [
                    SPNEGO_MechType(oid=ASN1_OID(oid)) for oid in supported_ssps
                ]
                self.supported_mechtypes.sort(
                    key=lambda x: SPNEGOSSP._PREF_ORDER.index(x.oid.val)
                )
            else:
                self.supported_mechtypes = force_supported_mechtypes
            super(SPNEGOSSP.CONTEXT, self).__init__()

        # Passthrough attributes and functions

        def clifailure(self):
            self.sub_context.clifailure()

        def __getattr__(self, attr):
            try:
                return object.__getattribute__(self, attr)
            except AttributeError:
                return getattr(self.sub_context, attr)

        def __setattr__(self, attr, val):
            try:
                return object.__setattr__(self, attr, val)
            except AttributeError:
                return setattr(self.sub_context, attr, val)

        # Passthrough the flags property

        @property
        def flags(self):
            if self.sub_context:
                return self.sub_context.flags
            return GSS_C_FLAGS(0)

        @flags.setter
        def flags(self, x):
            if not self.sub_context:
                return
            self.sub_context.flags = x

        def __repr__(self):
            return "SPNEGOSSP[%s]" % repr(self.sub_context)

    _MECH_ALIASES = {
        # Kerberos has 2 ssps
        "1.2.840.48018.1.2.2": "1.2.840.113554.1.2.2",
        "1.2.840.113554.1.2.2": "1.2.840.48018.1.2.2",
    }

    # This is the order Windows chooses. We mimic it for plausibility
    _PREF_ORDER = [
        "1.2.840.48018.1.2.2",  # MS KRB5
        "1.2.840.113554.1.2.2",  # Kerberos 5
        "1.3.6.1.4.1.311.2.2.30",  # NEGOEX
        "1.3.6.1.4.1.311.2.2.10",  # NTLM
    ]

    def __init__(self, ssps, **kwargs):
        self.supported_ssps = {x.oid: x for x in ssps}
        # Apply MechTypes aliases
        for ssp in ssps:
            if ssp.oid in self._MECH_ALIASES:
                self.supported_ssps[self._MECH_ALIASES[ssp.oid]] = self.supported_ssps[
                    ssp.oid
                ]
        self.force_supported_mechtypes = kwargs.pop("force_supported_mechtypes", None)
        super(SPNEGOSSP, self).__init__(**kwargs)

    def _extract_gssapi(self, Context, x):
        status, otherMIC, rawToken = None, None, False
        # Extract values from GSSAPI
        if isinstance(x, GSSAPI_BLOB):
            x = x.innerToken
        if isinstance(x, SPNEGO_negToken):
            x = x.token
        if hasattr(x, "mechTypes"):
            Context.requested_mechtypes = x.mechTypes
            Context.negotiated_mechtype = None
        if hasattr(x, "supportedMech") and x.supportedMech is not None:
            Context.negotiated_mechtype = x.supportedMech
        if hasattr(x, "mechListMIC") and x.mechListMIC:
            otherMIC = GSSAPI_BLOB_SIGNATURE(x.mechListMIC.value.val)
        if hasattr(x, "_mechListMIC") and x._mechListMIC:
            otherMIC = GSSAPI_BLOB_SIGNATURE(x._mechListMIC.value.val)
        if hasattr(x, "negResult"):
            status = x.negResult
        try:
            x = x.mechToken
        except AttributeError:
            try:
                x = x.responseToken
            except AttributeError:
                # No GSSAPI wrapper (windows fallback). Remember this for answer
                rawToken = True
        if isinstance(x, SPNEGO_Token):
            x = x.value
        if Context.requested_mechtypes:
            try:
                cls = _mechDissector[
                    (
                        Context.negotiated_mechtype or Context.requested_mechtypes[0]
                    ).oid.val  # noqa: E501
                ]
            except KeyError:
                cls = conf.raw_layer
            if isinstance(x, ASN1_STRING):
                x = cls(x.val)
            elif isinstance(x, conf.raw_layer):
                x = cls(x.load)
        return x, status, otherMIC, rawToken

    def NegTokenInit2(self):
        """
        Server-Initiation of GSSAPI/SPNEGO.
        See [MS-SPNG] sect 3.2.5.2
        """
        Context = self.CONTEXT(
            self.supported_ssps,
            force_supported_mechtypes=self.force_supported_mechtypes,
        )
        return (
            Context,
            GSSAPI_BLOB(
                innerToken=SPNEGO_negToken(
                    token=SPNEGO_negTokenInit(mechTypes=Context.supported_mechtypes)
                )
            ),
        )

        # NOTE: NegoEX has an effect on how the SecurityContext is
        # initialized, as detailed in [MS-AUTHSOD] sect 3.3.2
        # But the format that the Exchange token uses appears not to
        # be documented :/

        # resp.SecurityBlob.innerToken.token.mechTypes.insert(
        #     0,
        #     # NEGOEX
        #     SPNEGO_MechType(oid="1.3.6.1.4.1.311.2.2.30"),
        # )
        # resp.SecurityBlob.innerToken.token.mechToken = SPNEGO_Token(
        #     value=negoex_token
        # )  # noqa: E501

    def GSS_WrapEx(self, Context, *args, **kwargs):
        # Passthrough
        return Context.ssp.GSS_WrapEx(Context.sub_context, *args, **kwargs)

    def GSS_UnwrapEx(self, Context, *args, **kwargs):
        # Passthrough
        return Context.ssp.GSS_UnwrapEx(Context.sub_context, *args, **kwargs)

    def GSS_GetMICEx(self, Context, *args, **kwargs):
        # Passthrough
        return Context.ssp.GSS_GetMICEx(Context.sub_context, *args, **kwargs)

    def GSS_VerifyMICEx(self, Context, *args, **kwargs):
        # Passthrough
        return Context.ssp.GSS_VerifyMICEx(Context.sub_context, *args, **kwargs)

    def LegsAmount(self, Context: CONTEXT):
        return 4

    def _common_spnego_handler(self, Context, IsClient, val=None, req_flags=None):
        if Context is None:
            # New Context
            Context = SPNEGOSSP.CONTEXT(
                self.supported_ssps,
                req_flags=req_flags,
                force_supported_mechtypes=self.force_supported_mechtypes,
            )
            if IsClient:
                Context.requested_mechtypes = Context.supported_mechtypes

        # Extract values from GSSAPI token
        status, MIC, otherMIC, rawToken = 0, None, None, False
        if val:
            val, status, otherMIC, rawToken = self._extract_gssapi(Context, val)

        # If we don't have a SSP already negotiated, check for requested and available
        # SSPs and find a common one.
        if Context.ssp is None:
            if Context.negotiated_mechtype is None:
                if Context.requested_mechtypes:
                    # Find a common SSP
                    try:
                        Context.negotiated_mechtype = next(
                            x
                            for x in Context.requested_mechtypes
                            if x in Context.supported_mechtypes
                        )
                    except StopIteration:
                        # no common mechanisms
                        raise ValueError("No common SSP mechanisms !")
                    # Check whether the selected SSP was the one preferred by the client
                    if (
                        Context.negotiated_mechtype != Context.requested_mechtypes[0]
                        and val
                    ):
                        Context.first_choice = False
                # No SSPs were requested. Use the first available SSP we know.
                elif Context.supported_mechtypes:
                    Context.negotiated_mechtype = Context.supported_mechtypes[0]
                else:
                    raise ValueError("Can't figure out what SSP to use")
            # Set Context.ssp to the object matching the chosen SSP type.
            Context.ssp = self.supported_ssps[Context.negotiated_mechtype.oid.val]

        if not Context.first_choice:
            # The currently provided token is not for this SSP !
            # Typically a client opportunistically starts with Kerberos, including
            # its APREQ, and we want to use NTLM. We add one round trip
            Context.state = SPNEGOSSP.STATE.FIRST
            Context.first_choice = True  # reset to not come here again.
            tok, status = None, GSS_S_CONTINUE_NEEDED
        else:
            # The currently provided token is for this SSP !
            # Pass it to the sub ssp, with its own context
            if IsClient:
                (
                    Context.sub_context,
                    tok,
                    status,
                ) = Context.ssp.GSS_Init_sec_context(
                    Context.sub_context,
                    val=val,
                    req_flags=Context.req_flags,
                )
            else:
                Context.sub_context, tok, status = Context.ssp.GSS_Accept_sec_context(
                    Context.sub_context, val=val
                )
            # Check whether client or server says the specified mechanism is not valid
            if status == GSS_S_BAD_MECH:
                # Mechanism is not usable. Typically the Kerberos SPN is wrong
                to_remove = [Context.negotiated_mechtype.oid.val]
                # If there's an alias (for the multiple kerberos oids, also include it)
                if Context.negotiated_mechtype.oid.val in SPNEGOSSP._MECH_ALIASES:
                    to_remove.append(
                        SPNEGOSSP._MECH_ALIASES[Context.negotiated_mechtype.oid.val]
                    )
                # Drop those unusable mechanisms from the supported list
                for x in list(Context.supported_mechtypes):
                    if x.oid.val in to_remove:
                        Context.supported_mechtypes.remove(x)
                # Re-calculate negotiated mechtype
                try:
                    Context.negotiated_mechtype = next(
                        x
                        for x in Context.requested_mechtypes
                        if x in Context.supported_mechtypes
                    )
                except StopIteration:
                    # no common mechanisms
                    raise ValueError("No common SSP mechanisms after GSS_S_BAD_MECH !")
                # Start again.
                Context.state = SPNEGOSSP.STATE.CHANGESSP
                Context.ssp = None  # Reset the SSP
                Context.sub_context = None  # Reset the SSP context
                if IsClient:
                    # Call ourselves again for the client to generate a token
                    return self._common_spnego_handler(Context, True, None)
                else:
                    # Return nothing but the supported SSP list
                    tok, status = None, GSS_S_CONTINUE_NEEDED

        if rawToken:
            # No GSSAPI wrapper (fallback)
            return Context, tok, status

        # Client success
        if IsClient and tok is None and status == GSS_S_COMPLETE:
            return Context, None, status

        # Map GSSAPI codes to SPNEGO
        if status == GSS_S_COMPLETE:
            negResult = 0  # accept_completed
        elif status == GSS_S_CONTINUE_NEEDED:
            negResult = 1  # accept_incomplete
        else:
            negResult = 2  # reject

        # GSSAPI-MIC
        if Context.ssp and Context.ssp.canMechListMIC(Context.sub_context):
            # The documentation on mechListMIC wasn't clear, so note that:
            # - The mechListMIC that the client sends is computed over the
            #   list of mechanisms that it requests.
            # - the mechListMIC that the server sends is computed over the
            #   list of mechanisms that the client requested.
            # Yes, this does indeed mean that NegTokenInit2 added by [MS-SPNG]
            # is NOT protected. That's not necessarily an issue, since it was
            # optional in most cases, but it's something to keep in mind.
            if otherMIC is not None:
                # Check the received MIC if any
                if IsClient:  # from server
                    Context.ssp.verifyMechListMIC(
                        Context,
                        otherMIC,
                        mechListMIC(Context.supported_mechtypes),
                    )
                else:  # from client
                    Context.ssp.verifyMechListMIC(
                        Context,
                        otherMIC,
                        mechListMIC(Context.requested_mechtypes),
                    )
            # Then build our own MIC
            if IsClient:  # client
                if negResult == 0:
                    # Include MIC for the last packet. We could add a check
                    # here to only send the MIC when required (when preferred ssp
                    # isn't chosen)
                    MIC = Context.ssp.getMechListMIC(
                        Context,
                        mechListMIC(Context.supported_mechtypes),
                    )
            else:  # server
                MIC = Context.ssp.getMechListMIC(
                    Context,
                    mechListMIC(Context.requested_mechtypes),
                )

        if IsClient:
            if Context.state == SPNEGOSSP.STATE.FIRST:
                # First client token
                spnego_tok = SPNEGO_negToken(
                    token=SPNEGO_negTokenInit(mechTypes=Context.supported_mechtypes)
                )
                if tok:
                    spnego_tok.token.mechToken = SPNEGO_Token(value=tok)
            else:
                # Subsequent client tokens
                spnego_tok = SPNEGO_negToken(  # GSSAPI_BLOB is stripped
                    token=SPNEGO_negTokenResp(
                        supportedMech=None,
                        negResult=None,
                    )
                )
                if tok:
                    spnego_tok.token.responseToken = SPNEGO_Token(value=tok)
                if Context.state == SPNEGOSSP.STATE.CHANGESSP:
                    # On renegotiation, include the negResult and chosen mechanism
                    spnego_tok.token.negResult = negResult
                    spnego_tok.token.supportedMech = Context.negotiated_mechtype
        else:
            spnego_tok = SPNEGO_negToken(  # GSSAPI_BLOB is stripped
                token=SPNEGO_negTokenResp(
                    supportedMech=None,
                    negResult=negResult,
                )
            )
            if Context.state in [SPNEGOSSP.STATE.FIRST, SPNEGOSSP.STATE.CHANGESSP]:
                # Include the supportedMech list if this is the first thing we do
                # or a renegotiation.
                spnego_tok.token.supportedMech = Context.negotiated_mechtype
            if tok:
                spnego_tok.token.responseToken = SPNEGO_Token(value=tok)
        # Apply MIC if available
        if MIC:
            spnego_tok.token.mechListMIC = SPNEGO_MechListMIC(
                value=ASN1_STRING(MIC),
            )
        if (
            IsClient and Context.state == SPNEGOSSP.STATE.FIRST
        ):  # Client: after the first packet, specifying 'SPNEGO' is implicit.
            # Always implicit for the server.
            spnego_tok = GSSAPI_BLOB(innerToken=spnego_tok)
        # Not the first token anymore
        Context.state = SPNEGOSSP.STATE.NORMAL
        return Context, spnego_tok, status

    def GSS_Init_sec_context(
        self, Context: CONTEXT, val=None, req_flags: Optional[GSS_C_FLAGS] = None
    ):
        return self._common_spnego_handler(Context, True, val=val, req_flags=req_flags)

    def GSS_Accept_sec_context(self, Context: CONTEXT, val=None):
        return self._common_spnego_handler(Context, False, val=val, req_flags=0)

    def GSS_Passive(self, Context: CONTEXT, val=None):
        if Context is None:
            # New Context
            Context = SPNEGOSSP.CONTEXT(self.supported_ssps)
            Context.passive = True

        # Extraction
        val, status, _, rawToken = self._extract_gssapi(Context, val)

        if val is None and status == GSS_S_COMPLETE:
            return Context, None

        # Just get the negotiated SSP
        if Context.negotiated_mechtype:
            mechtype = Context.negotiated_mechtype
        elif Context.requested_mechtypes:
            mechtype = Context.requested_mechtypes[0]
        elif rawToken and Context.supported_mechtypes:
            mechtype = Context.supported_mechtypes[0]
        else:
            return None, GSS_S_BAD_MECH
        try:
            ssp = self.supported_ssps[mechtype.oid.val]
        except KeyError:
            return None, GSS_S_BAD_MECH

        if Context.ssp is not None:
            # Detect resets
            if Context.ssp != ssp:
                Context.ssp = ssp
                Context.sub_context = None
        else:
            Context.ssp = ssp

        # Passthrough
        Context.sub_context, status = Context.ssp.GSS_Passive(Context.sub_context, val)

        return Context, status

    def GSS_Passive_set_Direction(self, Context: CONTEXT, IsAcceptor=False):
        Context.ssp.GSS_Passive_set_Direction(
            Context.sub_context, IsAcceptor=IsAcceptor
        )

    def MaximumSignatureLength(self, Context: CONTEXT):
        return Context.ssp.MaximumSignatureLength(Context.sub_context)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
Very partial RPC definitions for the following interfaces:
- logon (v1.0): 12345678-1234-ABCD-EF00-01234567CFFB
"""

from enum import IntEnum
import uuid

from scapy.fields import StrFixedLenField
from scapy.layers.dcerpc import (
    register_dcerpc_interface,
    DceRpcOp,
    NDRConfVarStrNullFieldUtf16,
    NDRFullPointerField,
    NDRInt3264EnumField,
    NDRIntField,
    NDRPacket,
    NDRPacketField,
)


class PNETLOGON_CREDENTIAL(NDRPacket):
    fields_desc = [StrFixedLenField("data", "", length=8)]


class PNETLOGON_AUTHENTICATOR(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRPacketField("Credential", PNETLOGON_CREDENTIAL(), PNETLOGON_CREDENTIAL),
        NDRIntField("Timestamp", 0),
    ]


class NetrServerReqChallenge_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("PrimaryName", "")),
        NDRConfVarStrNullFieldUtf16("ComputerName", ""),
        NDRPacketField("ClientChallenge", PNETLOGON_CREDENTIAL(), PNETLOGON_CREDENTIAL),
    ]


class NetrServerReqChallenge_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("ServerChallenge", PNETLOGON_CREDENTIAL(), PNETLOGON_CREDENTIAL),
        NDRIntField("status", 0),
    ]


class NETLOGON_SECURE_CHANNEL_TYPE(IntEnum):
    NullSecureChannel = 0
    MsvApSecureChannel = 1
    WorkstationSecureChannel = 2
    TrustedDnsDomainSecureChannel = 3
    TrustedDomainSecureChannel = 4
    UasServerSecureChannel = 5
    ServerSecureChannel = 6
    CdcServerSecureChannel = 7


class NetrServerAuthenticate3_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("PrimaryName", "")),
        NDRConfVarStrNullFieldUtf16("AccountName", ""),
        NDRInt3264EnumField("SecureChannelType", 0, NETLOGON_SECURE_CHANNEL_TYPE),
        NDRConfVarStrNullFieldUtf16("ComputerName", ""),
        NDRPacketField(
            "ClientCredential", PNETLOGON_CREDENTIAL(), PNETLOGON_CREDENTIAL
        ),
        NDRIntField("NegotiateFlags", 0),
    ]


class NetrServerAuthenticate3_Response(NDRPacket):
    fields_desc = [
        NDRPacketField(
            "ServerCredential", PNETLOGON_CREDENTIAL(), PNETLOGON_CREDENTIAL
        ),
        NDRIntField("NegotiateFlags", 0),
        NDRIntField("AccountRid", 0),
        NDRIntField("status", 0),
    ]


LOGON_OPNUMS = {
    4: DceRpcOp(NetrServerReqChallenge_Request, NetrServerReqChallenge_Response),
    26: DceRpcOp(NetrServerAuthenticate3_Request, NetrServerAuthenticate3_Response),
}
register_dcerpc_interface(
    name="logon",
    uuid=uuid.UUID("12345678-1234-ABCD-EF00-01234567CFFB"),
    version="1.0",
    opnums=LOGON_OPNUMS,
)



# SPDX-License-Identifier: GPL-2.0-or-later
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
[MS-DRSR] Directory Replication Service (DRS) Remote Protocol
"""

import uuid
from scapy.packet import Packet
from scapy.fields import LEIntField, FlagsField, UUIDField, UTCTimeField

from scapy.layers.msrpce.raw.ms_drsr import UUID
from scapy.layers.msrpce.raw.ms_drsr import *  # noqa: F403,F401

# [MS-DRSR] sect 5.39 DRS_EXTENSIONS_INT


class DRS_EXTENSIONS_INT(Packet):
    fields_desc = [
        FlagsField(
            "dwFlags",
            0,
            -32,
            {
                0x00000001: "BASE",
                0x00000002: "ASYNCREPL",
                0x00000004: "REMOVEAPI",
                0x00000008: "MOVEREQ_V2",
                0x00000010: "GETCHG_DEFLATE",
                0x00000020: "DCINFO_V1",
                0x00000040: "RESTORE_USN_OPTIMIZATION",
                0x00000080: "ADDENTRY",
                0x00000100: "KCC_EXECUTE",
                0x00000200: "ADDENTRY_V2",
                0x00000400: "LINKED_VALUE_REPLICATION",
                0x00000800: "DCINFO_V2",
                0x00001000: "INSTANCE_TYPE_NOT_REQ_ON_MOD",
                0x00002000: "CRYPTO_BIND",
                0x00004000: "GET_REPL_INFO",
                0x00008000: "STRONG_ENCRYPTION",
                0x00010000: "DCINFO_VFFFFFFFF",
                0x00020000: "TRANSITIVE_MEMBERSHIP",
                0x00040000: "ADD_SID_HISTORY",
                0x00080000: "POST_BETA3",
                0x00100000: "GETCHGREQ_V5",
                0x00200000: "GETMEMBERSHIPS2",
                0x00400000: "GETCHGREQ_V6",
                0x00800000: "NONDOMAIN_NCS",
                0x01000000: "GETCHGREQ_V8",
                0x02000000: "GETCHGREPLY_V5",
                0x04000000: "GETCHGREPLY_V6",
                0x08000000: "WHISTLER_BETA3",
                0x10000000: "W2K3_DEFLATE",
                0x20000000: "GETCHGREQ_V10",
                0x40000000: "R2",
                0x80000000: "R3",
            },
        ),
        UUIDField("SiteObjGuid", None, uuid_fmt=UUIDField.FORMAT_LE),
        LEIntField("Pid", 0),
        UTCTimeField("dwReplEpoch", None, fmt="<I"),
        FlagsField(
            "dwFlagsExt",
            0,
            -32,
            {
                0x00000001: "ADAM",
                0x00000002: "LH_BETA2",
                0x00000004: "RECYCLE_BIN",
                0x00000100: "GETCHGREPLY_V9",
                0x00000400: "RPC_CORRELATIONID_1",
            },
        ),
        UUIDField("ConfigObjGuid", None, uuid_fmt=UUIDField.FORMAT_LE),
        LEIntField("dwExtCaps", 0),
    ]


# [MS-DRSR] sect 5.138 NTDSAPI_CLIENT_GUID

NTDSAPI_CLIENT_GUID = UUID(uuid.UUID("{e24d201a-4fd6-11d1-a3da-0000f875ae0d}").bytes_le)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
Very partial RPC definitions for the following interfaces:
- srvsvc (v3.0): 4B324FC8-1670-01D3-1278-5A47BF6EE188
"""

import uuid

from scapy.fields import StrFixedLenField
from scapy.layers.dcerpc import (
    register_dcerpc_interface,
    DceRpcOp,
    NDRConfPacketListField,
    NDRConfVarStrNullFieldUtf16,
    NDRFullPointerField,
    NDRIntField,
    NDRPacket,
    NDRPacketField,
    NDRUnionField,
)


class LPSHARE_INFO_1(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("shi1_netname", ""), deferred=True
        ),
        NDRIntField("shi1_type", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("shi1_remark", ""), deferred=True
        ),
    ]


class SHARE_INFO_1_CONTAINER(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("EntriesRead", None, size_of="Buffer"),
        NDRFullPointerField(
            NDRConfPacketListField(
                "Buffer",
                [LPSHARE_INFO_1()],
                LPSHARE_INFO_1,
                count_from=lambda pkt: pkt.EntriesRead,
            ),
            deferred=True,
        ),
    ]


class LPSHARE_ENUM_STRUCT(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("Level", 0),
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "ShareInfo",
                            SHARE_INFO_1_CONTAINER(),
                            SHARE_INFO_1_CONTAINER,
                        ),
                        deferred=True,
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                ),
            ],
            StrFixedLenField("ShareInfo", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
    ]


class NetrShareEnum_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("ServerName", "")),
        NDRPacketField("InfoStruct", LPSHARE_ENUM_STRUCT(), LPSHARE_ENUM_STRUCT),
        NDRIntField("PreferedMaximumLength", 0),
        NDRFullPointerField(NDRIntField("ResumeHandle", 0)),
    ]


class NetrShareEnum_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("InfoStruct", LPSHARE_ENUM_STRUCT(), LPSHARE_ENUM_STRUCT),
        NDRIntField("TotalEntries", 0),
        NDRFullPointerField(NDRIntField("ResumeHandle", 0)),
        NDRIntField("status", 0),
    ]


class NetrShareGetInfo_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("ServerName", "")),
        NDRConfVarStrNullFieldUtf16("NetName", ""),
        NDRIntField("Level", 0),
    ]


class NetrShareGetInfo_Response(NDRPacket):
    fields_desc = [
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField("ShareInfo", LPSHARE_INFO_1(), LPSHARE_INFO_1)
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 1),
                        (lambda _, val: val.tag == 1),
                    ),
                ),
            ],
            StrFixedLenField("ShareInfo", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
        NDRIntField("status", 0),
    ]


class LPSERVER_INFO_101(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("sv101_platform_id", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("sv101_name", ""), deferred=True
        ),
        NDRIntField("sv101_version_major", 0),
        NDRIntField("sv101_version_minor", 0),
        NDRIntField("sv101_type", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("sv101_comment", ""), deferred=True
        ),
    ]


class NetrServerGetInfo_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("ServerName", "")),
        NDRIntField("Level", 0),
    ]


class NetrServerGetInfo_Response(NDRPacket):
    fields_desc = [
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "ServerInfo", LPSERVER_INFO_101(), LPSERVER_INFO_101
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 101),
                        (lambda _, val: val.tag == 101),
                    ),
                ),
            ],
            StrFixedLenField("ServerInfo", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
        NDRIntField("status", 0),
    ]


SRVSVC_OPNUMS = {
    15: DceRpcOp(NetrShareEnum_Request, NetrShareEnum_Response),
    16: DceRpcOp(NetrShareGetInfo_Request, NetrShareGetInfo_Response),
    21: DceRpcOp(NetrServerGetInfo_Request, NetrServerGetInfo_Response),
}
register_dcerpc_interface(
    name="srvsvc",
    uuid=uuid.UUID("4B324FC8-1670-01D3-1278-5A47BF6EE188"),
    version="3.0",
    opnums=SRVSVC_OPNUMS,
)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
Very partial RPC definitions for the following interfaces:
- IObjectExporter (v0.0): 99fcfec4-5260-101b-bbcb-00aa0021347a
"""


import uuid

from scapy.fields import StrFixedLenField
from scapy.layers.dcerpc import (
    NDRPacket,
    DceRpcOp,
    NDRConfPacketListField,
    NDRConfStrLenField,
    NDRConfStrLenFieldUtf16,
    NDRFullPointerField,
    NDRIntField,
    NDRPacketField,
    NDRShortField,
    register_dcerpc_interface,
)


# Basic ORPC structures


class COMVERSION(NDRPacket):
    ALIGNMENT = (2, 2)
    fields_desc = [NDRShortField("MajorVersion", 0), NDRShortField("MinorVersion", 0)]


class GUID(NDRPacket):
    ALIGNMENT = (4, 4)
    fields_desc = [
        NDRIntField("Data1", 0),
        NDRShortField("Data2", 0),
        NDRShortField("Data3", 0),
        StrFixedLenField("Data4", "", length=8),
    ]


class ORPC_EXTENT(NDRPacket):
    ALIGNMENT = (4, 8)
    DEPORTED_CONFORMANTS = ["data"]
    fields_desc = [
        NDRPacketField("id", GUID(), GUID),
        NDRIntField("size", 0),
        NDRConfStrLenField(
            "data",
            "",
            size_is=lambda pkt: ((pkt.size + 7) & (~7)),
            conformant_in_struct=True,
        ),
    ]


class ORPC_EXTENT_ARRAY(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("size", 0),
        NDRIntField("reserved", 0),
        NDRFullPointerField(
            NDRConfPacketListField(
                "extent",
                [ORPC_EXTENT()],
                ORPC_EXTENT,
                size_is=lambda pkt: ((pkt.size + 1) & (~1)),
            ),
            deferred=True,
        ),
    ]


class ORPCTHIS(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRPacketField("version", COMVERSION(), COMVERSION),
        NDRIntField("flags", 0),
        NDRIntField("reserved1", 0),
        NDRPacketField("cid", GUID(), GUID),
        NDRFullPointerField(
            NDRPacketField("extensions", ORPC_EXTENT_ARRAY(), ORPC_EXTENT_ARRAY),
            deferred=True,
        ),
    ]


class MInterfacePointer(NDRPacket):
    ALIGNMENT = (4, 8)
    DEPORTED_CONFORMANTS = ["abData"]
    fields_desc = [
        NDRIntField("ulCntData", None, size_of="abData"),
        NDRConfStrLenField(
            "abData", "", size_is=lambda pkt: pkt.ulCntData, conformant_in_struct=True
        ),
    ]


class ORPCTHAT(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("flags", 0),
        NDRFullPointerField(
            NDRPacketField("extensions", ORPC_EXTENT_ARRAY(), ORPC_EXTENT_ARRAY),
            deferred=True,
        ),
    ]


class DUALSTRINGARRAY(NDRPacket):
    ALIGNMENT = (4, 8)
    DEPORTED_CONFORMANTS = ["aStringArray"]
    fields_desc = [
        NDRShortField("wNumEntries", None, size_of="aStringArray"),
        NDRShortField("wSecurityOffset", 0),
        NDRConfStrLenFieldUtf16(
            "aStringArray",
            "",
            size_is=lambda pkt: pkt.wNumEntries,
            conformant_in_struct=True,
        ),
    ]


# A few RPCs


class ServerAlive_Request(NDRPacket):
    fields_desc = []


class ServerAlive_Response(NDRPacket):
    fields_desc = [NDRIntField("status", 0)]


class ServerAlive2_Request(NDRPacket):
    fields_desc = []


class ServerAlive2_Response(NDRPacket):
    fields_desc = [
        NDRPacketField("pComVersion", COMVERSION(), COMVERSION),
        NDRFullPointerField(
            NDRPacketField("ppdsaOrBindings", DUALSTRINGARRAY(), DUALSTRINGARRAY)
        ),
        NDRIntField("pReserved", 0),
        NDRIntField("status", 0),
    ]


IOBJECTEXPORTER_OPNUMS = {
    3: DceRpcOp(ServerAlive_Request, ServerAlive_Response),
    5: DceRpcOp(ServerAlive2_Request, ServerAlive2_Response),
}
register_dcerpc_interface(
    name="IObjectExporter",
    uuid=uuid.UUID("99fcfec4-5260-101b-bbcb-00aa0021347a"),
    version="0.0",
    opnums=IOBJECTEXPORTER_OPNUMS,
)



# SPDX-License-Identifier: GPL-2.0-only
# This file is part of Scapy
# See https://scapy.net/ for more information
# Copyright (C) Gabriel Potter

"""
Very partial RPC definitions for the following interfaces:
- wkssvc (v1.0): 6BFFD098-A112-3610-9833-46C3F87E345A
"""

from enum import IntEnum
import uuid

from scapy.fields import StrFixedLenField
from scapy.layers.dcerpc import (
    register_dcerpc_interface,
    DceRpcOp,
    NDRConfPacketListField,
    NDRConfVarStrLenFieldUtf16,
    NDRConfVarStrNullFieldUtf16,
    NDRFullPointerField,
    NDRInt3264EnumField,
    NDRIntField,
    NDRPacket,
    NDRPacketField,
    NDRShortField,
    NDRUnionField,
)


class LPWKSTA_INFO_100(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("wki100_platform_id", 0),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("wki100_computername", ""), deferred=True
        ),
        NDRFullPointerField(
            NDRConfVarStrNullFieldUtf16("wki100_langroup", ""), deferred=True
        ),
        NDRIntField("wki100_ver_major", 0),
        NDRIntField("wki100_ver_minor", 0),
    ]


class NetrWkstaGetInfo_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("ServerName", "")),
        NDRIntField("Level", 0),
    ]


class NetrWkstaGetInfo_Response(NDRPacket):
    fields_desc = [
        NDRUnionField(
            [
                (
                    NDRFullPointerField(
                        NDRPacketField(
                            "WkstaInfo", LPWKSTA_INFO_100(), LPWKSTA_INFO_100
                        )
                    ),
                    (
                        (lambda pkt: getattr(pkt, "Level", None) == 100),
                        (lambda _, val: val.tag == 100),
                    ),
                ),
            ],
            StrFixedLenField("WkstaInfo", "", length=0),
            align=(4, 8),
            switch_fmt=("L", "L"),
        ),
        NDRIntField("status", 0),
    ]


class NET_COMPUTER_NAME_TYPE(IntEnum):
    NetPrimaryComputerName = 0
    NetAlternateComputerNames = 1
    NetAllComputerNames = 2
    NetComputerNameTypeMax = 3


class PUNICODE_STRING(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRShortField("Length", None, size_of="Buffer", adjust=lambda _, x: (x * 2)),
        NDRShortField(
            "MaximumLength", None, size_of="Buffer", adjust=lambda _, x: (x * 2)
        ),
        NDRFullPointerField(
            NDRConfVarStrLenFieldUtf16(
                "Buffer", "", length_from=lambda pkt: (pkt.Length // 2)
            ),
            deferred=True,
        ),
    ]


class PNET_COMPUTER_NAME_ARRAY(NDRPacket):
    ALIGNMENT = (4, 8)
    fields_desc = [
        NDRIntField("EntryCount", None, size_of="ComputerNames"),
        NDRFullPointerField(
            NDRConfPacketListField(
                "ComputerNames",
                [PUNICODE_STRING()],
                PUNICODE_STRING,
                count_from=lambda pkt: pkt.EntryCount,
            ),
            deferred=True,
        ),
    ]


class NetrEnumerateComputerNames_Request(NDRPacket):
    fields_desc = [
        NDRFullPointerField(NDRConfVarStrNullFieldUtf16("ServerName", "")),
        NDRInt3264EnumField("NameType", 0, NET_COMPUTER_NAME_TYPE),
        NDRIntField("Reserved", 0),
    ]


class NetrEnumerateComputerNames_Response(NDRPacket):
    fields_desc = [
        NDRFullPointerField(
            NDRPacketField(
                "ComputerNames", PNET_COMPUTER_NAME_ARRAY(), PNET_COMPUTER_NAME_ARRAY
            )
        ),
        NDRIntField("status", 0),
    ]


WKSSVC_OPNUMS = {
    0: DceRpcOp(NetrWkstaGetInfo_Request, NetrWkstaGetInfo_Response),
    30: DceRpcOp(
        NetrEnumerateComputerNames_Request, NetrEnumerateComputerNames_Response
    ),
}
register_dcerpc_interface(
    name="wkssvc",
    uuid=uuid.UUID("6BFFD098-A112-3610-9833-46C3F87E345A"),
    version="1.0",
    opnums=WKSSVC_OPNUMS,
)
